#
# File:   jtag2mem.tcl
# Date:   09-Dec-05
# Author: I. Chuang
#
# TCL script to be used with Xilinx's xtclsh to read and write data from
# PC to a Xilinx FPGA via the JTAG interface.  The FPGA code should inclue
# the jtag2mem.v verilog module.
#
# This script allows the FPGA memory to be read and written, using the 
# following syntax:
#
#   xtclsh jtag2mem.tcl read <addr> <nbytes> <optional file_name>
#   xtclsh jtag2mem.tcl write <addr> <optional file_name>
#
# Data is read from and written to standard input/output (the terminal).
# For a write, type the hex data in after starting the script, or
# read from a file.  For a read, the data appears on stdout; you
# can also have it written to a file.
#
# The file should contain hex bytes separated by whitespace
# 
# Note that the address should be a hex number, with exactly the number 
# of bits necessary for the jtag2mem.v implementation you have.  For the 
# sample configuration use three digits.
#
#---------------------------------------------------------------------------
#
# For example, with the jtag2mem_6111 example bitfile programmed into the 
# labkit, try:
#
#  xtclsh jtag2mem.tcl read 3ff 1
#
#      reads one byte of the memory at address 3ff (should give switch values)
#
#  xtclsh jtag2mem.tcl write 000
#
#      lets you write to memory starting at location 000 (should set LEDs)
#
#  xtclsh jtag2mem.tcl write 000 charfill.dat 
#
#      fills up the character memory with the alphabet, in upper and lower case
#
#  xtclsh jtag2mem.tcl read 000 100 out.dat 
#
#      reads 100 bytes of memory starting from addr 000, writing to out.dat
#
# Note that to run jtag2mem.tcl, you will need various Xilinx libraries,
# including versions of libTclJtag10.dll and libJtagcomm10.dll appropriate
# for the release of ISE which you are running.  Also needed is the modified
# tcl jtag interface code, mytcljtag.tcl, and the file deviceid.lst
#
# For the 6.111 labkits, be sure that jtag_open uses cable type 
# xilinx_parallel4 and not xilinx_platformusb in the code below.
# Also, for the parallel port cables, you must quite IMPACT before
# jtag2mem can run properly (only one program can use the cable at a time).
#
#----------------------------------------------------------------------------
# 
#  This file is part of jtag2mem, (c) 2005 Isaac Chuang <ichuang@mit.edu>
# 
#  jtag2mem is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or (at your
#  option) any later version.
# 
#  jtag2mem is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this code; if not, write to the Free Software Foundation,
#  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#----------------------------------------------------------------------------
# JTAG procedures

proc set_u1 {handle} {
    # Shift the USER1 Instruction (b00010) into the Instruction Register
    jtag_shiftir      $handle -buffer "010000" -endstate RTI -device 1
}

proc set_u2 {handle} {
    # Shift the USER2 Instruction (b00011) into the Instruction Register
    jtag_shiftir      $handle -buffer "110000" -endstate RTI -device 1
}

proc send_ur {handle urbits} {
    set bout [jtag_shiftdr $handle -buffer $urbits -endstate RTI -device 1]
    # puts "mode bout is $bout"
    return $bout
}

proc setup_jtag {} {
    # setup jtag communications
    #set handle        [jtag_open]
#    set handle        [jtag_open type=xilinx_platformusb]
    set handle        [jtag_open type=xilinx_parallel4 port=lpt1 ]
    jtag_lock         $handle
    jtag_autodetect   $handle
    return $handle
}

proc end_jtag {handle} {
    jtag_unlock    $handle
    jtag_close     $handle
}

#----------------------------------------------------------------------------
# jtag2mem procedures

proc set_mode {handle modebits} {
    set_u1 $handle
    # write to user data register 1 to set jtag2mem mode
    send_ur $handle $modebits
    # puts "mode bout is $bout"
}

proc xfer_data {handle databits} {
    set_u2 $handle
    # write to user data register 2 to set data/addr
    return [send_ur $handle $databits]
}

proc set_addr {handle addr} {
    # put into CMD_LDADR=4 mode 
    set_mode $handle "0010"
    set addrbits [Hex2Bin $addr]
    xfer_data $handle $addrbits
}

#----------------------------------------------------------------------------
# read/write main routines

proc do_write {addr infile hexdat} {

    set handle [setup_jtag]

    # set address
    set_addr $handle $addr
    set_addr $handle $addr

    # set write data mode (CMD_WRDAT = "0100")
    set_mode $handle "0100"

    # write data one word at a time
    set_u2 $handle 
    set i 0
    foreach arg $hexdat {
	set bin [Hex2Bin $arg]
	# puts "dat $i = $arg : $bin"
	send_ur $handle $bin
	# puts "bout is $bout"
	incr i
    }
    puts "$i bytes written to memory"

    # finish up
    set_u2 $handle
    end_jtag $handle
}

proc do_read {addr nbytes outfile} {

    set handle [setup_jtag]

    # set address
    set_addr $handle $addr
    set_addr $handle $addr

    # set read data mode (CMD_WRDAT = "0000")
    set_mode $handle "0000"

    if {[string equal $outfile stdout]} {
	puts "------------------------------------------------------------"
	puts "Data read from memory:"
    }

    # read data one word at a time
    set_u2 $handle 
    for {set i 1} {$i <= $nbytes} {incr i} {
	# puts "dat $i = $arg : $bin"
        set_u2 $handle
	set bout [send_ur $handle "00000000"]
	set hex [Bin2Hex $bout]
	puts -nonewline $outfile "$hex "
	if {[expr $i % 26]==0} {
	    puts $outfile ""
	}
    }

    if {[string equal $outfile stdout]} {
	puts ""
    } else {
	puts $outfile ""
	close $outfile
    }

    # finish up
    set_u2 $handle
    end_jtag $handle
}

#----------------------------------------------------------------------------
# main code

source mytcljtag.tcl
set argn 0

puts "------------------------------------------------------------"

# get command ("read" or "write")
set cmd [lindex $argv $argn]
set argn [expr $argn + 1]
puts "cmd = $cmd"

# get address
set addr [lindex $argv $argn]
set argn [expr $argn + 1]
puts "address = $addr"

if {[string equal -nocase $cmd "read"]} {
    set nbytes [lindex $argv $argn]
    set argn [expr $argn + 1]
    puts "nbytes = $nbytes"

    if {$argc > $argn} {
	set fname [lindex $argv $argn]
	set argn [expr $argn + 1]
	puts "fname = $fname"
	set infile [open $fname w]
    } else {
	set infile stdout
    }

    puts "------------------------------------------------------------"
    do_read $addr $nbytes $infile
}

if {[string equal -nocase $cmd "write"]} {

    if {$argc > $argn} {
	set fname [lindex $argv $argn]
	set argn [expr $argn + 1]
	puts "fname = $fname"
	set infile [open $fname r]
	# read data from file
	set hexdat {}
	while { [gets $infile line] >= 0 } {
	    set hexdat [concat $hexdat [split $line]]
	}
    } else {
	set infile stdin
	puts "Type one line of hex data (separate bytes by space):"
	gets $infile line
	set hexdat [split $line]
    }

    set number [llength $hexdat]
    puts "Number of hex digits: $number"
    puts "------------------------------------------------------------"
    close $infile

    do_write $addr $infile $hexdat
}

    

